local L = LibStub('AceLocale-3.0'):GetLocale('TimeToDie', true)

local TimeToDie = LibStub('AceAddon-3.0'):NewAddon('TimeToDie', 'AceEvent-3.0', 'AceConsole-3.0')
_G.TimeToDie = TimeToDie

local SML = LibStub("LibSharedMedia-3.0")

------------------------------
--		Initialization		--
------------------------------

local events = {target = 'PLAYER_TARGET_CHANGED', focus = 'PLAYER_FOCUS_CHANGED'}
local healing = '|cff00ff00'..SHOW_COMBAT_HEALING ..'|r'

local UnitIsFriend = UnitIsFriend
local UnitHealth = UnitHealth
local GetTime = GetTime
local UnitHealthMax = UnitHealthMax
local format = format

local defaults = {
	profile = {
		enemy = true,
		target = false, -- false := 'target', true := 'focus'
		tot = false,

		frame = true,
		locked = false,
		font = defaultFont,
		size = 24,
		outline = '',
		r = 1,
		g = .82,
		b = 0,
		strata = 'LOW',
		justify = 'CENTER',

		p1 = 'CENTER',
		p2 = 'CENTER',
		x = 0,
		y = 0,
	}
}

function TimeToDie:OnInitialize()
	local db = LibStub('AceDB-3.0'):New('TimeToDieDB', defaults, 'Default')
	self.db = db
	local RegisterCallback = db.RegisterCallback
	RegisterCallback(self, 'OnProfileChanged', 'OnEnable')
	RegisterCallback(self, 'OnProfileCopied', 'OnEnable')
	RegisterCallback(self, 'OnProfileReset', 'OnEnable')

	self:RegisterOptions()
end

function TimeToDie:OnEnable()
	local RegisterEvent = self.RegisterEvent
	RegisterEvent(self, 'PLAYER_TARGET_CHANGED', 'OnChange')
	RegisterEvent(self, 'PLAYER_FOCUS_CHANGED', 'OnChange')
	RegisterEvent(self, 'UNIT_HEALTH')

	local profile = self.db.profile
	self:ApplySettings(profile)

	if profile.frame then
		self:UpdateFrame(profile)
	elseif self.frame then
		self.frame:Hide()
	end
end

function TimeToDie:OnDisable()
	if self.frame then self.frame:Hide() end
	self:DisableEventFrame()
end

-------------------------------
--		Event Functions		--
-------------------------------

local health0, time0 -- initial health and time point
local mhealth, mtime -- current midpoint
local profileTarget -- := profile.target and 'focus' or 'target'
local currentTarget -- profileTarget or profileTagert..'target'

function TimeToDie:UNIT_HEALTH(event, unit, health)  --health is not passed from blizz, and hopefully never will be
	if unit ~= currentTarget then return end

	health = health or UnitHealth(unit)

	if health == UnitHealthMax(unit) then
		self.dataobj.text, health0, time0, mhealth, mtime = nil
		return
	end

	local time = GetTime()

	if not health0 then
		health0, time0 = health, time
		return
	end

	if mhealth then
		mhealth, mtime = (mhealth + health) / 2, (mtime + time) / 2
	else
		mhealth, mtime = health, time
	end

	local dhealth = mhealth - health0

	local res
	if dhealth ~= 0 then
		time = health * (time0 - mtime) / dhealth
		if time < 0 then
			res = healing
			health0, time0, mhealth, mtime = nil
		elseif time > 60 then
			res = format('%d:%02d', time/60, time%60)
		else
			res = format('%d', time%60)
		end
	else
		res = '0s'
	end
	self.dataobj.text = res
end

function TimeToDie:OnChange(event)
	local target = profileTarget
	if events[target] ~= event then return end

	self.dataobj.text, health0, time0, mhealth, mtime = nil

	local profile = self.db.profile

	if profile.enemy and UnitIsFriend('player', target) then
		if profile.tot then
			target = target..'target'
			if not UnitExists('player', target) or UnitIsFriend('player', target) then
				target = nil
			end
		else
			target = nil
		end
	end
	currentTarget = target
   if not target and self.eventFrame then self.eventFrame:Hide() end -- stops error in OnUpdate
   return target -- for UNIT_TARGET
end

local oldhealth = nil
local totalElapsed = 0
local function OnUpdate(self, elapsed)
	totalElapsed = totalElapsed + elapsed
	if totalElapsed < .1 then return end
	totalElapsed = 0

	local health = UnitHealth(currentTarget)
	if oldhealth ~= health then
		oldhealth = health
		TimeToDie:UNIT_HEALTH('UNIT_HEALTH', currentTarget, health)
	end
end

--unit := target, player, or nil (focus)
local function UNIT_TARGET(self, event, unit)
	if unit == 'player' then
      unit = 'target'
   elseif event == 'PLAYER_FOCUS_CHANGED' then
      unit = 'focus'
   end
	if unit == profileTarget then -- unit is now either 'target' or 'focus'
      local target = TimeToDie:OnChange(events[unit])
      if target and target ~= unit then --target == unit..'target'
         self:Show()
      else
         self:Hide()
      end
	end
end

-- profile.enemey and profile.tot = true
-- only use if target is friendly
function TimeToDie:EnableEventFrame()
	local frame = self.eventFrame
	if not frame then
		frame = CreateFrame('Frame')
		self.eventFrame = frame
		frame.OnEvent = UNIT_TARGET
		frame:SetScript('OnEvent', UNIT_TARGET)
		frame.OnUpdate = OnUpdate
		frame:SetScript('OnUpdate', OnUpdate)
	end
   frame:Hide() -- turn off OnUpdate when changing options
	frame:RegisterEvent('UNIT_TARGET') -- covers player, targettarget, and focustarget
	frame:RegisterEvent('PLAYER_FOCUS_CHANGED')  -- covers focus
	--make a call in case targets are already set when enabled
	UNIT_TARGET(frame, nil, profileTarget)
end

function TimeToDie:DisableEventFrame()
	local frame = self.eventFrame
	if frame then
		frame:Hide()
		frame:UnregisterAllEvents()
	end
end

function TimeToDie:ApplySettings(profile)
	timer, health0, time0, mhealth, mtime = nil

	profileTarget = profile.target and 'focus' or 'target'

	if profile.enemy and profile.tot then
		self:EnableEventFrame()
	else
		self:DisableEventFrame()
		self:OnChange(events[profileTarget])  -- this is called from EnableEventFrame, so need one here
	end
end

function TimeToDie:UpdateFrame(profile)
	local frame = self.frame
	if not frame then
		frame = CreateFrame('Frame', 'TimeToDieFrame', UIParent, 'TimeToDieFrameTemplate')
		self.frame = frame
	end
	local bg, text = frame.bg, frame.text

	frame:ClearAllPoints()
	frame:SetPoint(profile.p1, UIParent, profile.p2, profile.x, profile.y)

	if profile.frame then
		frame:Show()
	else
		frame:Hide()
		profile.locked = true
	end

	if profile.locked then
		bg:Hide()
		frame:EnableMouse(false)
		self.dataobj.text = nil
	else
		bg:Show()
		frame:EnableMouse(true)
		self.dataobj.text = 'TimeToDie'
	end

	text:SetFont(SML:Fetch('font', profile.font), profile.size, profile.outline)
	text:SetTextColor(profile.r, profile.g, profile.b)

	frame:SetFrameStrata(profile.strata)
	text:SetJustifyH(profile.justify)
end

TimeToDie.dataobj = LibStub:GetLibrary('LibDataBroker-1.1'):NewDataObject('TimeToDie', {
	type = 'data source',
	text = 'TTD',
	label = 'TTD',
	icon = 'Interface\\Icons\\Spell_Holy_BorrowedTime',
})

function TimeToDie:Debug(...)
	if not self.debugging then return end
	if not IsAddOnLoaded('Blizzard_DebugTools') then LoadAddOn('Blizzard_DebugTools') end
	EventTraceFrame:Show()
	EventTraceFrame_OnEvent(EventTraceFrame, ...)
end
